import functools
from datetime import datetime
from os import chdir
from shutil import copy
import yaml
from ase import Atoms
from ase.calculators.vasp import Vasp
from ase.io import read
from htsct.utils.tools import Pipe
from htsct.io.vaspIO import write_incar, write_kpoints, write_optcell
from pathlib import Path


def task_log_decorator(task_type):
    """
    类装饰器，用于记录类中函数执行时间
    :param task_type:工作类型
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            st = datetime.now()
            rt = func(self, *args, **kwargs)
            et = datetime.now()
            time_span = et - st
            _yaml_fp = Path(self.log_path)
            _yaml_fp.parent.mkdir(parents=True, exist_ok=True)
            with open(_yaml_fp, 'a+') as fd:
                yaml.dump(
                    {st: {'path': str(self.root), "task type": task_type, "time used": str(time_span)}},
                    fd)
            return rt

        return wrapper

    return decorator


class Calculator:
    def __init__(self, root, log_path, command):
        self.optimized_atoms = None
        self.root = Path(root).absolute()
        if not self.root.exists():
            self.root.mkdir()
        self.scf_folder = self.root / "scf"
        self.initial_atoms = self.read_struct_from_file()
        self.log_path = Path(log_path)
        self.command = command

    @staticmethod
    def set_calc(atoms, directory, command, **kwargs):
        calc = Vasp()
        calc.set(atoms=atoms, directory=directory, command=command, **kwargs)
        calc.get_potential_energy()
        return calc

    def read_struct_from_file(self):
        initial_atoms = None
        for file in self.root.iterdir():
            if ".cif" in file.name:
                initial_atoms = read(self.root / file)
            elif "POSCAR" in file.name:
                initial_atoms = read(self.root / "POSCAR")
        if not initial_atoms:
            raise Exception("Can not read structure from file(.cif|POSCAR)")
        return initial_atoms

    @task_log_decorator("struct")
    def struct(self, **kwargs):
        """
        Structure optimization calculator.
        There have some special kwargs:
        optcell=1D-a|1D-b|1D-c|2D|3D,default is 3D.
        :param kwargs:ase.calculators.vasp.Vasp(**kwargs)
        :return:instance of ase.calculators.vasp.Vasp
        """
        if "optcell" in kwargs.keys():
            optcell = kwargs.pop("optcell")
            write_optcell(self.root, optcell)
        if not isinstance(self.initial_atoms, Atoms):
            raise ValueError(f"The input structure is wrong!-->initial_atoms:{self.initial_atoms}")
        return self.set_calc(atoms=self.initial_atoms, command=self.command, directory=self.root, **kwargs)

    @task_log_decorator("scf")
    def scf(self, **kwargs):
        """
        Scf calculator.
        :param kwargs:ase.calculators.vasp.Vasp(**kwargs)
        :return:instance of ase.calculators.vasp.Vasp
        """
        if not self.scf_folder.exists():
            self.scf_folder.mkdir()
        self.optimized_atoms = read(self.root / "CONTCAR")
        (self.root / "WAVECAR").replace(self.scf_folder / "WAVECAR")
        return self.set_calc(atoms=self.optimized_atoms, command=self.command, directory=self.scf_folder, **kwargs)


class BandCalc(Calculator):
    def __init__(self, root, log_path, command):
        super(BandCalc, self).__init__(root, log_path, command)
        self.band_path = self.root / "band"

    @task_log_decorator("band")
    def band(self, **kwargs):
        """
        Band calculator.
        :param kwargs:Vasp INCAR parameters
        """
        if not self.band_path.exists():
            self.band_path.mkdir()
        copy_list = [[self.scf_folder / "CHGCAR", self.band_path],
                     [self.scf_folder / "WAVECAR", self.band_path],
                     [self.scf_folder / "CONTCAR", self.band_path / "POSCAR"],
                     [self.scf_folder / "POTCAR", self.band_path]]
        for i in copy_list:
            copy(i[0], i[1])
        chdir(self.band_path)
        if "vaspkit" in kwargs.keys():
            vaspkit = kwargs.pop("vaspkit")
            if vaspkit == "1D":
                Pipe("vaspkit -task 301\n", self.band_path)
            elif vaspkit == "2D":
                Pipe("vaspkit -task 302\n", self.band_path)
            elif vaspkit == "3D":
                Pipe("vaspkit -task 303\n", self.band_path)
            try:
                copy(self.band_path / "KPATH.in", self.band_path / "KPOINTS")
            except FileNotFoundError as e:
                print(e)
                return -1
        else:
            write_kpoints(
                directory="KPOINTS",
                kpaths=kwargs.pop("kpaths"),
                kpts=kwargs.pop("kpts"),
                density=kwargs.pop("density")
            )
        write_incar(directory="INCAR", **kwargs)
        stdout = Pipe(self.command, self.band_path).stdout
        with open(self.band_path / "vasp.out", "w+") as tmp:
            tmp.write(stdout)
