import json
import re
import sqlite3
from pathlib import Path
from typing import Union
from textwrap import dedent
from ase.io import read
import matplotlib as mpl
from shutil import copy, rmtree
from htsct.constants import SYSTEM_LOG, CALC_FILES
from htsct.db.sqlite import update_db_by_id, write_db
from htsct.utils.tools import Pipe, md5_hex, matchList, matchManyList

mpl.use('Agg')  # 使得终端不显示图片


class VaspOutAnalysis:
    def __init__(self, fn=""):
        self.__filename = Path(fn)
        if "OUTCAR" != self.__filename.name:
            self.__filename = self.__filename / "OUTCAR"
        self.errors = []
        self.warnings = []

    def is_converge(self):
        with open(self.__filename) as fd:
            for line in fd:
                if "reached required accuracy" in line:
                    return True
            return False

    def error_collection(self):
        errors = ""
        with open(self.__filename) as fd:
            for line in fd:
                if "Error" in line or "error" in line or "failed" in line:
                    errors += line
            if errors:
                return errors
            return ""

    def warning_collection(self):
        warnings = ""
        with open(self.__filename) as fd:
            for line in fd:
                if "WARNING" in line or "Warning" in line:
                    warnings += line
            if warnings:
                return warnings
            return ""


def structOptimizationAnalysis(struct_path):
    vaspOutAnalysis = VaspOutAnalysis(struct_path)
    if not vaspOutAnalysis.is_converge():
        with SYSTEM_LOG.open("w+") as fd:
            err = vaspOutAnalysis.error_collection()
            warn = vaspOutAnalysis.warning_collection()
            fd.write(dedent(f"""{err}\n{warn}
                Structure cannot be converged!!Directory is {struct_path}!"""))
        raise Exception(f"结构优化未收敛，计算中止。中止计算的文件夹为{struct_path}")


class DataExtractor:
    def __init__(self, root, db_dir, table_name):
        self.root = Path(root)
        self.conn = sqlite3.connect(db_dir, timeout=10)
        self.table_name = table_name
        self.unique_id = ""
        self.struct_data = None
        self.scf_data = None
        self.band_data = None
        self.target_path = Path(CALC_FILES) / self.root.name

    @staticmethod
    def __structureDataExtractor(struct_dir: Union[str, Path]):
        """提取结构优化数据
        """
        struct_dir = Path(struct_dir)
        poscar = struct_dir / "POSCAR"
        contcar = struct_dir / "CONTCAR"
        oszicar = struct_dir / "OSZICAR"
        formula = read(poscar).get_chemical_formula()
        oszicar_content = oszicar.read_text(encoding="utf-8")
        line = matchList(oszicar_content.split("\n")[-2::], r".*?F=(.*?)E0=(.*?)d E.*?mag=(.*)")
        last, E0, mag = re.match(r".*?F=(.*?)E0=(.*?)d E.*?mag=(.*)", line).groups()
        return {
            "unique_id": md5_hex(poscar),
            "Formula": formula,
            "Path": struct_dir.__str__(),
            "POSCAR": poscar.read_text(encoding="utf-8"),
            "CONTCAR": contcar.read_text(encoding="utf-8"),
            "Last step energy": float(last),
            "E0": float(E0),
            "mag": float(mag)
        }

    @staticmethod
    def __scfDataExtractor(scf_dir):
        """提取scf数据
        """
        scf_dir = Path(scf_dir)
        lines1 = Pipe("vaspkit -task 426\n", scf_dir).stdout.splitlines(True)
        lines2 = Pipe("grep E-fermi OUTCAR\n", scf_dir).stdout.splitlines(True)
        return {
            "vacuum_level": float(matchList(lines1, r"\s*?Vacuum-Level\s*?\(eV\)").split()[-1]),
            "e_fermi": float(matchList(lines2, r"\s*?E-fermi\s*?").split()[2])
        }

    @staticmethod
    def __bandDataExtractor(band_dir):
        """提取能带计算数据
        """
        band_dir = Path(band_dir).absolute()
        Pipe("vaspkit -task 211\n", band_dir)
        BAND_GAP = (band_dir / "BAND_GAP").read_text(encoding="utf-8").splitlines()
        patterns = [
            r"\s*?Band\s*?Gap",  # band_gap
            r"\s*?Eigenvalue of VBM\s*?",  # _vbm
            r"\s*?Eigenvalue of CBM\s*?",  # _cbm
            r"\s*?Fermi Energy\s*?",  # fermi_energy
            r"\s*?Location of VBM\s*?",  # vbm_position
            r"\s*?Location of CBM\s*?"  # cbm_position
        ]
        band_gap, _vbm, _cbm, fermi_energy, vbm_position, cbm_position = matchManyList(BAND_GAP, patterns)
        return {
            "band_gap": float(band_gap.split()[-1]),
            "_vbm": float(_vbm.split()[-1]),
            "_cbm": float(_cbm.split()[-1]),
            "fermi_energy": float(fermi_energy.split()[-1]),
            "vbm_position": [float(vbm_position.split()[-3]),
                             float(vbm_position.split()[-2]),
                             float(vbm_position.split()[-1])],
            "cbm_position": [float(cbm_position.split()[-3]),
                             float(cbm_position.split()[-2]),
                             float(cbm_position.split()[-1])]
        }

    @staticmethod
    def calcData(data: dict):
        data["work_function"] = data["vacuum_level"] - data["e_fermi"]
        return {
            "cbm": data["_cbm"] - data["fermi_energy"] - data["work_function"],
            "vbm": data["_vbm"] - data["fermi_energy"] - data["work_function"]
        }

    def structDataWriter(self):
        structOptimizationAnalysis(self.root)  # 判断结构是否收敛，若不收敛则抛出异常
        self.struct_data = DataExtractor.__structureDataExtractor(self.root)
        self.unique_id = self.struct_data["unique_id"]
        self.copyFiles("struct")
        self.struct_data.update({"target_path": self.target_path.__str__()})
        write_db(self.conn, self.table_name, self.struct_data)

    def scfDataWriter(self):
        self.scf_data = DataExtractor.__scfDataExtractor(self.root / "scf")
        self.copyFiles("scf")
        update_db_by_id(self.conn, self.table_name, self.scf_data, self.unique_id)

    def bandDataWriter(self):
        self.band_data = DataExtractor.__bandDataExtractor(self.root / "band")
        all_data = {**self.struct_data, **self.scf_data, **self.band_data}
        calc_data = self.calcData(all_data)
        with open(self.root / "band" / "Band_Position_Info", "w+") as fd:
            fd.write(json.dumps({**all_data, **calc_data}, indent=2, sort_keys=True))
        self.copyFiles("band")
        update_db_by_id(self.conn, self.table_name, {**self.band_data, **calc_data}, self.unique_id)

    def copyFiles(self, folder_name):
        target_path = self.target_path / folder_name
        src_path = self.root / folder_name if folder_name != "struct" else self.root
        if target_path.exists():
            rmtree(target_path)
        target_path.mkdir(parents=True, exist_ok=True)
        for f in src_path.iterdir():
            if f.is_file():
                copy(f.__str__(), target_path.__str__())

    def clearFiles(self):
        rmtree(self.root)
