import functools
import os
import re
import sys
from copy import deepcopy
from io import BytesIO
from multiprocessing import Process
from os import listdir
from os.path import dirname, abspath, exists, sep
from pathlib import Path, PurePath
from shutil import copy
from typing import Union

import yaml
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMainWindow, QFileDialog, QApplication, QWidget, QComboBox, QLineEdit, QCheckBox
from PyQt5.uic import loadUi
from ase.gui.gui import GUI
from ase.io import read
from ase.io.cif import parse_cif_ase
from typing.io import IO

from htsct.core.ribbon import Ribbon

NameOrFile = Union[str, PurePath, IO]


class ASEGUI(GUI):
    def __init__(self, parent=None, atoms=None, repeat_images=(3, 1, 1), *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(atoms, (str, PurePath, IO)):
            atoms = read(atoms)
        self.parent = parent
        self.new_atoms(atoms)
        self.images.repeat_images(repeat_images)
        self.set_frame()
        self.toggle_show_bonds(self)

    def show(self):
        if self.parent:
            self.parent.pushButton_showStructure.setText("Close")
        self.run()

    def close(self):
        self.exit()

    def exit(self, event=None):
        super().exit(event)
        if self.parent:
            self.parent.pushButton_showStructure.setText("Show")
            self.parent.ase_gui_status = False
            del self.parent.ase_gui
            self.parent.ase_gui = None


class LoadFile(QWidget):
    def __init__(self, parent):
        super(LoadFile, self).__init__()
        loadUi("loadfile.ui", self)
        self.parent = parent
        self.bindSlots()
        self.ase_gui = None
        self.ase_gui_status = False

    def bindSlots(self):
        self.pushButton_3.clicked.connect(lambda: self.open_file("lineEdit_2d_path"))
        self.pushButton_4.clicked.connect(lambda: self.open_file("lineEdit_ribbon_path"))
        self.lineEdit_ribbon_path.textChanged.connect(
            lambda: self.addItems("comboBox_file_name", self.lineEdit_ribbon_path.text()))
        self.comboBox_file_name.currentIndexChanged.connect(self.getInfo)
        self.pushButton_5.clicked.connect(self.pushButton_next)
        self.lineEdit_2d_path.textChanged.connect(self.setParent_root)
        self.lineEdit_source_fn.textChanged.connect(self.setParentStructureFile)
        self.pushButton_showStructure.clicked.connect(self.showStructure)
        self.lineEdit_milerIndex.textChanged.connect(self.setParentMilerIndex)
        self.lineEdit_width.textChanged.connect(self.setParentWidth)
        self.lineEdit_width_control.textChanged.connect(self.setParentWidthControl)
        self.pushButton_run.clicked.connect(self.parent.run_task)

    def setParentMilerIndex(self):
        milerIndex = self.lineEdit_milerIndex.text()
        if milerIndex:
            milerIndex = eval(milerIndex)
            a = ",".join(map(lambda _: str(_), milerIndex[0]))
            b = ",".join(map(lambda _: str(_), milerIndex[1]))
            c = ",".join(map(lambda _: str(_), milerIndex[2]))
            self.parent.set_vector(a, b, c)

    def setParentWidthControl(self):
        width_control = self.lineEdit_width_control.text()
        if width_control:
            width_control = eval(width_control)
            base = str(width_control[1])
            delta = str(width_control[2])
            self.parent.set_width_control(base, delta)

    def setParentWidth(self):
        width = self.lineEdit_width.text()
        if width:
            self.parent.lineEdit_fixedWidth.setText(width)

    def setParent_root(self):
        self.parent.lineEdit_fp.setText(self.lineEdit_2d_path.text())

    def setParentStructureFile(self):
        self.parent.comboBox_file.blockSignals(True)
        index = self.parent.comboBox_file.findText(self.lineEdit_source_fn.text())
        self.parent.comboBox_file.setCurrentIndex(index)
        self.parent.comboBox_file.blockSignals(False)

    def pushButton_next(self):
        lineEdit = ["lineEdit_fn_id", "lineEdit_source_fn", "lineEdit_milerIndex", "lineEdit_width",
                    "lineEdit_width_control"]
        for _ in lineEdit:
            if hasattr(self, _):
                getattr(self, _).clear()
        self.combobox_next("comboBox_file_name")
        if self.ase_gui_status:
            self.showStructure()
        if not self.ase_gui_status:
            self.showStructure()

    def open_file(self, lineEdit):
        fileDialog = QFileDialog(self, directory=".")
        fileDirectory = fileDialog.getExistingDirectory(self, "标题")
        if fileDirectory:
            if hasattr(self, lineEdit):
                getattr(self, lineEdit).setText(fileDirectory)

    def addItems(self, combobox, fp, f_type=".cif"):
        if hasattr(self, combobox):
            combobox = getattr(self, combobox)
            combobox.clear()
            if fp and exists(fp):
                cifs = [i for i in listdir(fp) if f_type in i]
                cifs = sorted(cifs)
                combobox.addItems(cifs)

    def setText(self, widget, text):
        if hasattr(self, widget):
            getattr(self, widget).setText(text)
            return True
        return False

    def getInfo(self):
        self.parent.comboBox_file.blockSignals(True)
        self.setText("lineEdit_fn_id", "")
        self.setText("lineEdit_milerIndex", "")
        self.setText("lineEdit_width", "")
        self.setText("lineEdit_width_control", "")
        self.setText("lineEdit_source_fn", "")
        cif_fp_2d = Path(self.lineEdit_2d_path.text())
        fn = self.comboBox_file_name.currentText()
        if not fn:
            return
        reg = re.compile(r"""
                (\d*?) # ID
                \..*?_\[
                (.*?) # miler index
                ]_\(
                (.*?) # width
                \)_?(?:\[
                (.*?) # width control
                ]|.*?)(_.*?)?.cif
            """, re.VERBOSE)
        reg2 = re.compile(r"(-?\d)")
        res = reg.findall(fn)
        if res and res[0][3]:
            res = res[0]
            mid = res[0]
            miler_index = [[int(j) for j in reg2.findall(i)] for i in res[1].split(',')]
            width = res[2]
            res3 = res[3].split(',')
            res3[2] = res3[2].strip("x")
            width_control = [int(i) for i in res3]
            self.setText("lineEdit_fn_id", mid)
            self.setText("lineEdit_milerIndex", repr(miler_index))
            self.setText("lineEdit_width", width)
            self.setText("lineEdit_width_control", repr(width_control))
            for i in listdir(cif_fp_2d):
                if mid in i.split(".")[0]:
                    self.setText("lineEdit_source_fn", i)
        self.parent.comboBox_file.blockSignals(False)

    def combobox_next(self, combobox):
        if hasattr(self, combobox):
            _ = getattr(self, combobox)
            index = _.currentIndex()
            count = _.count()
            if index < count - 1:
                _.setCurrentIndex(index + 1)
            else:
                _.setCurrentIndex(0)

    def showStructure(self):
        cif_fp_ribbon = Path(self.lineEdit_ribbon_path.text())
        fn = self.comboBox_file_name.currentText()
        current_file = cif_fp_ribbon / fn
        if not self.ase_gui:
            self.pushButton_showStructure.setText("Close")
            self.ase_gui_status = True
            if fn:
                self.ase_gui = ASEGUI(self, current_file)
                self.ase_gui.show()
        else:
            self.pushButton_showStructure.setText("Show")
            self.ase_gui_status = False
            self.ase_gui.close()
            del self.ase_gui
            self.ase_gui = None


class GUIProcess(Process):
    def __init__(self, data, *args, **kwargs):
        super(GUIProcess, self).__init__(*args, **kwargs)
        self._lineEdit_savePath = None
        self._comboBox_file = ""
        self._lineEdit_fp = ""
        self._lineEdit_repeat_image_c = None
        self._lineEdit_repeat_image_a = None
        self._lineEdit_repeat_image_b = None
        self._lineEdit_fixedWidth = None
        self._lineEdit_connectivity_opts = None
        self._lineEdit_maxWidth = None
        self._lineEdit_minWidth = None
        self._checkBox_showTarget = None
        self._checkBox_widthTest = None
        self._comboBox_delta = None
        self._comboBox_base = None
        self._lineEdit_top = None
        self._lineEdit_bottom = None
        self._comboBox_terminateAtom = None
        self._lineEdit_ribbon_tol = None
        self._lineEdit_vacuum_c = None
        self._lineEdit_vacuum_b = None
        self._comboBox_origo = None
        self._comboBox_vector_c = None
        self._comboBox_vector_b = None
        self._comboBox_vector_a = None
        for k, v in data.items():
            setattr(self, "_" + k, v)
        self._repeat_images = (self._lineEdit_repeat_image_a,
                               self._lineEdit_repeat_image_b,
                               self._lineEdit_repeat_image_c)
        self.fn = os.path.join(self._lineEdit_fp, self._comboBox_file)
        self.save_path = self._lineEdit_savePath if self._lineEdit_savePath else "./"

    def show_ribbon(self, ribbon):
        ase_gui = ASEGUI(None, ribbon, self._repeat_images)
        ase_gui.show()

    def run(self) -> None:
        fn = self.fn.split(sep)[-1].rstrip(".cif")
        for i in range(self._lineEdit_bottom, self._lineEdit_top):
            layer = 2 + self._comboBox_base + self._comboBox_delta * i
            ribbon = Ribbon.buildRibbon(self.fn, self._comboBox_vector_a, self._comboBox_vector_b,
                                        self._comboBox_vector_c,
                                        layer, self._comboBox_origo, self._lineEdit_vacuum_b,
                                        self._lineEdit_vacuum_c, self._lineEdit_ribbon_tol)
            ribbon_copy = deepcopy(ribbon)
            ribbon_copy.edge_del()
            ribbon_width = ribbon_copy.get_ribbon_width()
            ribbon.terminate(symbol=self._comboBox_terminateAtom, side=2,
                             connectivity_opts=self._lineEdit_connectivity_opts)
            print(f"当前纳米带宽度为：{ribbon_width}....")
            if self._lineEdit_fixedWidth:
                fixedWidth = self._lineEdit_fixedWidth
                ribbon_width = float(f"{ribbon_width:.2f}")
                if ribbon_width > fixedWidth:
                    break
                elif fixedWidth <= ribbon_width < fixedWidth + 0.01:
                    print(f"达到设置的固定宽度{fixedWidth}!!\n正在保存...")
                    if self.save_path:
                        if self._comboBox_terminateAtom:
                            save_ffp = self.save_path + sep + f'{fn}_[{self._comboBox_vector_a},{self._comboBox_vector_b},{self._comboBox_vector_c}]_({ribbon_width:.2f})_[{self._comboBox_origo},{self._comboBox_base},{self._comboBox_delta}x]_{self._comboBox_terminateAtom}.cif'
                        else:
                            save_ffp = self.save_path + sep + f'{fn}_[{self._comboBox_vector_a},{self._comboBox_vector_b},{self._comboBox_vector_c}]_({ribbon_width:.2f})_[{self._comboBox_origo},{self._comboBox_base},{self._comboBox_delta}x].cif'
                        with BytesIO() as b1:  # reload structure from cif
                            ribbon.write(b1, "cif")
                            b1.seek(0)
                            cif_block = parse_cif_ase(b1)
                            atoms = list(cif_block)[0].get_atoms()
                            self.show_ribbon(atoms)
                            atoms.write(save_ffp,"cif")
                        print(f"保存{save_ffp}成功")
                    break
                else:
                    print(f"未达到设置的固定宽度{fixedWidth}...")
                    continue
            elif self._lineEdit_minWidth < ribbon_width < self._lineEdit_maxWidth:
                if self._checkBox_showTarget:
                    self.show_ribbon(ribbon)
                print(f"当前纳米带符合条件：{ribbon_width},正在保存...", end="    ")
                if self.save_path:
                    save_ffp = self.save_path + sep + f'{fn}_[{self._comboBox_vector_a},{self._comboBox_vector_b},{self._comboBox_vector_c}]_({ribbon_width:.2f})_[{self._comboBox_origo},{self._comboBox_base},{self._comboBox_delta}x].cif'
                    ribbon.write(save_ffp, "cif")
                    print(f"保存{save_ffp}成功")
                break
            else:
                if self._checkBox_widthTest:
                    self.show_ribbon(ribbon)


def blockAllSignalDecorator(func):
    @functools.wraps(func)
    def wrapper(*args):
        objs = args[0].findChildren((QComboBox, QLineEdit))
        for obj in objs:
            obj.blockSignals(True)
        func(*args)
        for obj in objs:
            obj.blockSignals(False)

    return wrapper


class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        super(MainWindow, self).__init__(parent)
        loadUi("ribbonBuilder.ui", self)
        self.load_data()
        self.loadSettings("settings.yaml")
        self.flush_comboBox_file(self.getText(self, "lineEdit_fp"))
        self.bindSlots()
        self.tasks = []
        self.ase_gui_status = False  # 用于标志结构窗口是否打开，True代表打开
        self.ase_gui = None
        self.task_manage = None
        self.loadFileWidget = None

    def closeEvent(self, a0):
        self.saveSettings("settings.yaml")
        super(MainWindow, self).closeEvent(a0)

    @blockAllSignalDecorator
    def load_data(self):
        with open("config.yaml") as fd:
            data = yaml.load(fd, yaml.FullLoader)["UI Setup"]
            for k in data:
                if hasattr(self, k):
                    obj = getattr(self, k)
                    if isinstance(obj, QComboBox):
                        obj.addItems(data[k])
                    elif isinstance(obj, QLineEdit):
                        obj.setText(data[k])

            self.flush_comboBox_file(self.lineEdit_fp.text())

    def saveSettings(self, settings_fn: NameOrFile):
        data = {}
        lineEdits = self.findChildren(QLineEdit)
        for _ in lineEdits:
            data[_.objectName()] = _.text()
        comboBoxes = self.findChildren(QComboBox)
        for _ in comboBoxes:
            data[_.objectName()] = _.currentText()
        checkBoxes = self.findChildren(QCheckBox)
        for _ in checkBoxes:
            data[_.objectName()] = _.isChecked()
        with open(settings_fn, "w+") as fd:
            yaml.dump(data, fd)

    @blockAllSignalDecorator
    def loadSettings(self, settings_fn: NameOrFile):
        with open(settings_fn) as fd:
            data = yaml.load(fd, Loader=yaml.FullLoader)
        for obj_name, v in data.items():
            if hasattr(self, obj_name):
                obj = getattr(self, obj_name)
                if isinstance(obj, (QLineEdit, QComboBox)):
                    self.setText(self, obj_name, v)
                elif isinstance(obj, QCheckBox):
                    if v:
                        obj.setCheckState(Qt.Checked)
                    else:
                        obj.setCheckState(Qt.Unchecked)

    def bindSlots(self):
        self.pushButton_open.clicked.connect(self.open_file)
        self.pushButton_back.clicked.connect(
            lambda: self.lineEdit_fp.setText(dirname(self.lineEdit_fp.text())))
        self.lineEdit_fp.textChanged.connect(lambda: self.flush_comboBox_file(self.lineEdit_fp.text()))
        self.pushButton_next \
            .clicked.connect(self.next_structure)
        self.pushButton_save.clicked.connect(self.save_cif)
        self.pushButton_savePath.clicked.connect(
            lambda: self.lineEdit_savePath.setText(QFileDialog(self).getExistingDirectory(self, "标题")))
        self.pushButton_a_b.clicked.connect(self.exchange_a_b)
        self.pushButton_zigzag.clicked.connect(lambda: self.set_vector("1,1,0", "1,-1,0", "0,0,1"))
        self.pushButton_armchair.clicked.connect(lambda: self.set_vector("1,-1,0", "1,1,0", "0,0,1"))
        self.pushButton_H.clicked.connect(lambda: self.set_vector("1,0,0", "0,1,0", "0,0,1"))
        self.pushButton_V.clicked.connect(lambda: self.set_vector("0,1,0", "1,0,0", "0,0,1"))
        self.pushButton_1_2.clicked.connect(lambda: self.set_width_control("1", "2"))
        self.pushButton_2_2.clicked.connect(lambda: self.set_width_control("2", "2"))
        self.pushButton_2_3.clicked.connect(lambda: self.set_width_control("2", "3"))
        self.pushButton_showStructure.clicked.connect(self.show_structure)
        self.comboBox_file.currentIndexChanged.connect(self.structureChange)
        self.pushButton_run.clicked.connect(self.run_task)
        self.pushButton_stop.clicked.connect(self.stop_task)
        self.pushButton_loadFile.clicked.connect(self.loadFile)
        self.pushButton.clicked.connect(self.getData)
        self.actionReset_to_Default.triggered.connect(lambda: self.loadSettings("defaultSettings.yaml"))

    def exchange_a_b(self):
        text_a = self.getText(self, "comboBox_vector_a")
        text_b = self.getText(self, "comboBox_vector_b")
        self.setText(self, "comboBox_vector_a", text_b)
        self.setText(self, "comboBox_vector_b", text_a)

    def next_structure(self):
        index = self.comboBox_file.currentIndex()
        count = self.comboBox_file.count()
        if index < count - 1:
            self.comboBox_file.setCurrentIndex(index + 1)
        else:
            self.comboBox_file.setCurrentIndex(0)

    def save_cif(self):
        root = self.lineEdit_fp.text()  # 源文件所在目录
        fn = self.comboBox_file.currentText()  # 源文件名
        fp = os.path.join(root, fn)  # 源文件路径
        new_path = self.lineEdit_savePath.text()
        if not exists(new_path):
            os.makedirs(new_path)
        dest = os.path.join(dirname(root), f"{new_path}/{fn}")
        copy(fp, dest)

    @blockAllSignalDecorator
    def set_vector(self, a, b, c):
        self.setText(self, "comboBox_vector_a", a)
        self.setText(self, "comboBox_vector_b", b)
        self.setText(self, "comboBox_vector_c", c)

    def set_width_control(self, base, delta):
        self.setText(self, "comboBox_base", base)
        self.setText(self, "comboBox_delta", delta)

    def open_file(self):
        fileDialog = QFileDialog(self, directory=self.lineEdit_fp.text())
        fileDirectory = fileDialog.getExistingDirectory(self, "标题")
        if fileDirectory:
            self.lineEdit_fp.setText(fileDirectory)
            self.flush_comboBox_file(fileDirectory)

    def flush_comboBox_file(self, fn):
        self.comboBox_file.clear()
        if fn and exists(fn):
            cifs = [i for i in listdir(fn) if ".cif" in i]
            cifs = sorted(cifs)
            self.comboBox_file.addItems(cifs)

    def show_structure(self):
        root, fn = self.lineEdit_fp.text(), self.comboBox_file.currentText()
        if root and fn:
            current_file = os.path.join(root, fn)
            if self.ase_gui_status:
                self.ase_gui.close()
                self.ase_gui_status = False
            else:
                self.ase_gui_status = True
                self.ase_gui = ASEGUI(self, current_file, repeat_images=(3, 3, 1))
                self.ase_gui.show()

    def structureChange(self):
        if self.ase_gui_status:  # 如果原有结构界面已被打开，则关闭
            self.show_structure()
        if not self.ase_gui_status:  # 发现新的结构未被打开，则打开
            self.show_structure()

    @staticmethod
    def getText(parent: QWidget, widget: str):
        """Get text from QComboBox,QLineEdit"""
        if hasattr(parent, widget):
            obj = getattr(parent, widget)
            if isinstance(obj, QComboBox):
                return obj.currentText()
            elif isinstance(obj, QLineEdit):
                return obj.text()
            else:
                raise NotImplemented

    @staticmethod
    def setText(parent: QWidget, widget: str, text: str):
        if hasattr(parent, widget):
            obj = getattr(parent, widget)
            if isinstance(obj, QComboBox):
                index = obj.findText(text)
                if index >= 0:
                    obj.setCurrentIndex(index)
            elif isinstance(obj, QLineEdit):
                obj.setText(text)
            else:
                raise NotImplemented

    def getData(self):
        data = {}
        lineEdits = self.findChildren(QLineEdit)
        for _ in lineEdits:
            data[_.objectName()] = _.text()
        comboBoxes = self.findChildren(QComboBox)
        for _ in comboBoxes:
            data[_.objectName()] = _.currentText()
        checkBoxes = self.findChildren(QCheckBox)
        for _ in checkBoxes:
            data[_.objectName()] = _.isChecked()
        for k, v in data.items():
            try:
                if not isinstance(v, str):
                    continue
                if re.match(r"^-?([1-9]\d*\.\d*|0\.\d*[1-9]\d*|0?\.0+)$", v):  # match float
                    data[k] = float(v)
                elif re.match(r"^-?[1-9]\d*$", v):  # match int
                    data[k] = int(v)
                elif re.match(r"^(.*?,.*?,.*?|.*?,.*?)$", v):  # match xx,xx,xx
                    data[k] = eval("(" + data[k] + ")")
                elif re.match(r"^0$", v):
                    data[k] = 0
            except SyntaxError as e:
                print(e)
                continue
        return data

    def run_task(self):
        task = GUIProcess(self.getData())
        task.start()
        self.tasks.append(task)

    def stop_task(self):
        for task in self.tasks:
            task.terminate()

    def loadFile(self):
        self.loadFileWidget = LoadFile(self)
        self.loadFileWidget.show()

    def show(self):
        # 显示窗口
        super().show()
        desktop = QApplication.desktop()
        # 获取显示器分辨率大小
        screenRect = desktop.screenGeometry()
        height = screenRect.height()
        width = screenRect.width()
        self.resize(width * 0.65, height * 0.9)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    mainWindow = MainWindow()
    mainWindow.show()
    sys.exit(app.exec_())
